@@ -28,6 +28,7 @@ Resources = Dict[ResourceName, ResourceUrl]
 Specs = Dict[ResourceName, OpenApiSpec]
 ResourceInfo = Mapping[ResourceName, SpecPath]
 
+
 def _seq_but_not_str(obj: object) -> bool:
     """
     Determine if an object is a Sequence, i.e., has an iteratable type, but not a string, bytearray, etc.
@@ -37,76 +38,69 @@ def _seq_but_not_str(obj: object) -> bool:
     return isinstance(obj, Sequence) and not isinstance(obj, (str, bytes, bytearray))
 
 
-## currently the files spec is missing operationId's for some of its operations.
+# currently the files spec is missing operationId's for some of its operations.
 RESOURCES = {
-    'local': ['actors','authenticator', 'meta', 'files', 'sk', 'streams', 'systems', 'tenants','tokens'],
-    'tapipy':{
-        'actors': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-actors.yml',
-        'authenticator': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-authenticator.yml',
-        'meta': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-meta.yml',
-        'files': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-files.yml',
-        'sk': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-sk.yml',
-        'streams': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-streams.yml',
-        'systems': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-systems.yml',
-        'tenants': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-tenants.yml',
-        'tokens': 'https://raw.githubusercontent.com/tapis-project/tapipy/master/tapipy/resources/openapi_v3-tokens.yml'
+    'local':{
+        'actors': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-actors.yml",
+        'authenticator': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-authenticator.yml",
+        'meta': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-meta.yml",
+        'files': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-files.yml",
+        'sk': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-sk.yml",
+        'streams': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-streams.yml",
+        'systems': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-systems.yml",
+        'tenants': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-tenants.yml",
+        'tokens': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-tokens.yml",
     },
-    'master': {
-        'actors': 'https://raw.githubusercontent.com/TACC/abaco/master/docs/specs/openapi_v3.yml',               
-        'authenticator': 'https://raw.githubusercontent.com/tapis-project/authenticator/dev/service/resources/openapi_v3.yml',
-        'meta': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/master/meta-client/src/main/resources/metav3-openapi.yaml',
-        'files': 'https://raw.githubusercontent.com/tapis-project/tapis-files/master/api/src/main/resources/openapi.yaml',
-        'sk': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/master/security-client/src/main/resources/SKAuthorizationAPI.yaml',
-        'streams': 'https://raw.githubusercontent.com/tapis-project/streams-api/dev/service/resources/openapi_v3.yml',
-        'systems': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/master/systems-client/SystemsAPI.yaml',
-        'tenants': 'https://raw.githubusercontent.com/tapis-project/tenants-api/master/service/resources/openapi_v3.yml',
-        'tokens': 'https://raw.githubusercontent.com/tapis-project/tokens-api/master/service/resources/openapi_v3.yml'
+    'tapipy':{
+        'actors': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-actors.yml',
+        'authenticator': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-authenticator.yml',
+        'meta': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-meta.yml',
+        'files': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-files.yml',
+        'sk': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-sk.yml',
+        'streams': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-streams.yml',
+        'systems': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-systems.yml',
+        'tenants': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-tenants.yml',
+        'tokens': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-tokens.yml'
+    },
+    'prod': {
+        'actors': 'https://raw.githubusercontent.com/TACC/abaco/dev-v3/docs/specs/openapi_v3.yml',               
+        'authenticator': 'https://raw.githubusercontent.com/tapis-project/authenticator/prod/service/resources/openapi_v3.yml',
+        'meta': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/prod/meta-client/src/main/resources/metav3-openapi.yaml',
+        'files': 'https://raw.githubusercontent.com/tapis-project/tapis-files/prod/api/src/main/resources/openapi.yaml',
+        'sk': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/prod/security-client/src/main/resources/SKAuthorizationAPI.yaml',
+        'streams': 'https://raw.githubusercontent.com/tapis-project/streams-api/prod/service/resources/openapi_v3.yml',
+        'systems': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/prod/systems-client/SystemsAPI.yaml',
+        'tenants': 'https://raw.githubusercontent.com/tapis-project/tenants-api/prod/service/resources/openapi_v3.yml',
+        'tokens': 'https://raw.githubusercontent.com/tapis-project/tokens-api/prod/service/resources/openapi_v3.yml'
     },
     'dev': {
-        'actors': 'https://raw.githubusercontent.com/TACC/abaco/master/docs/specs/openapi_v3.yml',
+        'actors': 'https://raw.githubusercontent.com/TACC/abaco/dev-v3/docs/specs/openapi_v3.yml',
         'authenticator': 'https://raw.githubusercontent.com/tapis-project/authenticator/dev/service/resources/openapi_v3.yml',
-        'meta': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/master/meta-client/src/main/resources/metav3-openapi.yaml',
-        'files': 'https://raw.githubusercontent.com/tapis-project/tapis-files/master/api/src/main/resources/openapi.yaml',
+        'meta': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/dev/meta-client/src/main/resources/metav3-openapi.yaml',
+        'files': 'https://raw.githubusercontent.com/tapis-project/tapis-files/dev/api/src/main/resources/openapi.yaml',
         'sk': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/dev/security-client/src/main/resources/SKAuthorizationAPI.yaml',
         'streams': 'https://raw.githubusercontent.com/tapis-project/streams-api/dev/service/resources/openapi_v3.yml',
-        'systems': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/master/systems-client/SystemsAPI.yaml',
-        'tenants': 'https://raw.githubusercontent.com/tapis-project/tenants-api/master/service/resources/openapi_v3.yml',
-        'tokens': 'https://raw.githubusercontent.com/tapis-project/tokens-api/master/service/resources/openapi_v3.yml'
+        'systems': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/dev/systems-client/SystemsAPI.yaml',
+        'tenants': 'https://raw.githubusercontent.com/tapis-project/tenants-api/dev/service/resources/openapi_v3.yml',
+        'tokens': 'https://raw.githubusercontent.com/tapis-project/tokens-api/dev/service/resources/openapi_v3.yml'
     }
 }
 
 
-def get_specs_from_local_resource_defs(resources):
-    specs = {}
-    resources_dir = os.path.join(os.path.dirname(__file__), 'resources')
-    for r in resources:
-        try:
-            # for now, hardcode the paths; we could look these up based on a canonical URL once that is
-            # established.
-            spec_path = f'{resources_dir}/openapi_v3-{r}.yml'
-            spec_dict = yaml.load(open(spec_path, 'r'), Loader=yaml.FullLoader)
-            specs[r] = create_spec(spec_dict)
-        except:
-            pass
-    return specs
-
-
 def _get_specs(resources: Resources, spec_dir: str = None, download_latest_specs: bool = False) -> Specs:
     """
     Gets specs requested in resources.
     Will download any spec that is not already downloaded.
     """
-    # if no resources passed, return the specs corresponding to the resource folder --
-    if type(resources) == list:
-        return get_specs_from_local_resource_defs(resources)
     spec_dir = get_spec_dir(spec_dir)
     # Download and save specs if neccessary
-    download_and_pickle_spec_dicts(resources.values(), spec_dir=spec_dir, download_latest_specs=download_latest_specs)
+    download_and_pickle_spec_dicts(resources, spec_dir=spec_dir, download_latest_specs=download_latest_specs)
     # Load, unpickle, and create specs
     specs = unpickle_and_create_specs(resources, spec_dir=spec_dir)
     return specs
 
-def download_and_pickle_spec_dicts(url_list: list, spec_dir: str, download_latest_specs: bool) -> None:
+
+def download_and_pickle_spec_dicts(resources: Resources, spec_dir: str, download_latest_specs: bool) -> None:
     """
     Function that calls threads to download and pickle specs.
     Cuts wait time for 9 specs from 4s to 0.7s
@@ -114,13 +108,15 @@ def download_and_pickle_spec_dicts(url_list: list, spec_dir: str, download_lates
     # Get list of specs and check which we need to download
     # We download if the file name requested does not exist
     urls_to_download = []
-    for url in set(url_list):
+    for resource_name, url in resources.items():
+        if "local:" in url:
+            continue 
         _, full_spec_name, spec_path = get_file_info_from_url(url, spec_dir)
         if download_latest_specs:
-            urls_to_download.append([url, spec_path])
+            urls_to_download.append([resource_name, url, spec_path])
         else:
             if not full_spec_name in os.listdir(spec_dir):
-                urls_to_download.append([url, spec_path])
+                urls_to_download.append([resource_name, url, spec_path])
     
     # Set off some parallel processes cause it's quick.
     POOL_SIZE = os.environ.get('POOL_SIZE', 16)
@@ -129,13 +125,14 @@ def download_and_pickle_spec_dicts(url_list: list, spec_dir: str, download_lates
     pool.close()
     pool.join()
 
+
 def _thread_download_spec_dict(resource_info: ResourceInfo) -> None:
     """
     Function that multiprocessing pool calls to download and store pickled
     spec dicts. Gets spec dict, if it's valid, stores it at 'spec_path',
     else it does nothing.
     """
-    resource_url, spec_path = resource_info
+    resource_name, resource_url, spec_path = resource_info
     # Attempt to get spec from url
     response = requests.get(resource_url)
     if response.status_code == 200:
@@ -143,8 +140,8 @@ def _thread_download_spec_dict(resource_info: ResourceInfo) -> None:
             # Loads yaml into Python dictionary
             spec_dict = yaml.load(response.content, Loader=yaml.FullLoader)
         except Exception as e:
-            print(f'Got exception when attempting to load yaml for'
-                  f'"{spec_path}" resource; exception: {e}')
+            print(f'Got exception when attempting to load yaml for '
+                  f'"{resource_name}" resource; exception: {e}')
             return
         try:
             # Attempts to create spec from dict to ensure the spec is valid
@@ -153,21 +150,22 @@ def _thread_download_spec_dict(resource_info: ResourceInfo) -> None:
             test_spec_dict = copy.deepcopy(spec_dict)
             create_spec(test_spec_dict)
         except Exception as e:
-            print(f'Got exception when test creating spec for "{spec_path}" '
+            print(f'Got exception when test creating spec for "{resource_name}" '
                   f'resource; Spec probably not verifying; exception: {e}')
             return
         try:
             # Pickles and saves the spec dict to the spec_path atomically
             with atomic_write(f'{spec_path}', overwrite=True, mode='wb') as spec_file:
-                pickle.dump(spec_dict, spec_file, protocol=pickle.HIGHEST_PROTOCOL)
+                pickle.dump(spec_dict, spec_file, protocol=4)
         except Exception as e:
-            print(f'Got exception when attempting to pickle spec_dict for'
-                  f'"{spec_path}" resource; exception: {e}')
+            print(f'Got exception when attempting to pickle spec_dict for '
+                  f'"{resource_name}" resource; exception: {e}')
             return
     else:
-        raise KeyError(f'Error getting "{spec_path}" resource. URL: "{resource_url}".'
+        raise KeyError(f'Error getting "{resource_name}" resource. URL: "{resource_url}".'
                        f'Did not get 200, got the following back:\n{response.text}')
-    
+
+
 def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
     """
     Pickles loads a specifed spec_path and creates said spec.
@@ -176,6 +174,16 @@ def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
     specs = {}
     # Get resource path to point the unpickling at.
     for resource_name, url in resources.items():
+        if "local:" in url:
+            try:
+                spec_path = url.replace('local:', '').strip()
+                with open(spec_path, 'rb') as spec_file:
+                    spec_dict = yaml.load(spec_file, Loader=yaml.FullLoader)
+                specs.update({resource_name: create_spec(spec_dict)})
+            except Exception as e:
+                print(f'Error reading local "{resource_name}" resource. '
+                      f'Ensure path is absolute. e:{e}')
+            continue
         _, _, spec_path = get_file_info_from_url(url, spec_dir)
         try:
             # Unpickle and create_spec
@@ -183,10 +191,27 @@ def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
                 spec_dict = pickle.load(spec_file)
             specs.update({resource_name: create_spec(spec_dict)})
         except Exception as e:
-            print(f'Got exception trying to unpickle and create spec for'
-                  f'spec_path: "{spec_path}"; exception: {e}')
+            print(f'Got exception trying to unpickle and create spec for '
+                  f'spec_path: "{resource_name}"; exception: {e}')
+            # Using resource name to look if the resource exists in the
+            # tapipy prod resources. If it is found, we fall back and use that!
+            # If it isn't we just skip it.
+            print(f'Falling back to prod spec for "{resource_name}" resource')
+            if resource_name in RESOURCES['tapipy']:
+                url = RESOURCES['tapipy'][resource_name]
+                _, _, spec_path = get_file_info_from_url(url, spec_dir)
+                try:
+                    # Unpickle and create_spec
+                    with open(spec_path, 'rb') as spec_file:
+                        spec_dict = pickle.load(spec_file)
+                    specs.update({resource_name: create_spec(spec_dict)})
+                except Exception as e:
+                    print('Error opening tapipy prod spec. This is bad.')
+            else:
+                print(f'No "{resource_name}" spec was found to fallback on')
     return specs
 
+
 def update_spec_cache(resources: Resources = None, spec_dir: str = None) -> None:
     """
     Allows users to update specified specs in cache.
@@ -205,7 +230,8 @@ def update_spec_cache(resources: Resources = None, spec_dir: str = None) -> None
 
     spec_dir = get_spec_dir(spec_dir)
     download_and_pickle_spec_dicts(url_list, spec_dir=spec_dir, download_latest_specs=True)
-    
+
+
 def get_file_info_from_url(url: str, spec_dir: str):
     """
     Using a url string we create a file name to store said url contents.
@@ -221,6 +247,7 @@ def get_file_info_from_url(url: str, spec_dir: str):
     spec_path = f'{spec_dir}/{spec_name}.pickle'
     return spec_name, full_spec_name, spec_path
 
+
 def get_spec_dir(spec_dir: str):
     """
     Create a new directory for specs by copy Tapipy's to new dir
@@ -363,8 +390,14 @@ class Tapis(object):
         # Allows a user to specify which set of resources to pull from.
         # Only used when download_lastest_specs is used.
         self.resource_set = resource_set
+        if resource_set == 'prod':
+            self.resource_set = 'tapipy'
+        if self.resource_set == 'master':
+            raise errors.TapyClientConfigurationError("the master branch has been removed from tapis-project "
+                                                      "repositories; use 'prod' or 'tapipy' instead.")
         if not self.resource_set in RESOURCES.keys():
-            raise KeyError(f"'resource_set' must be one of {RESOURCES.keys()}, not {self.resource_set}.")
+            raise errors.TapyClientConfigurationError(f"'resource_set' must be one of {RESOURCES.keys()}, "
+                                                      f"not {self.resource_set}.")
 
         # If a custom spec dict is provided then the RESOURCES dict gets updated with it.
         # If any repeated fields are used, the RESOURCES fields are overwritten.
@@ -390,7 +423,7 @@ class Tapis(object):
         self.spec_dir = spec_dir
 
         # Uses module instantiated RESOURCE_SPECS if there are no changes to the specs. 
-        if self.custom_spec_dict or self.spec_dir or self.download_latest_specs or not self.resource_set == 'master':
+        if self.custom_spec_dict or self.spec_dir or self.download_latest_specs or not self.resource_set == 'tapipy':
             resource_specs = _get_specs(RESOURCES[resource_set], spec_dir=self.spec_dir, download_latest_specs=self.download_latest_specs)
         else:
             resource_specs = RESOURCE_SPECS
@@ -531,7 +564,7 @@ class Tapis(object):
                                                   target_site_id=target_site_id)
             except Exception as e:
                 raise errors.BaseTapyException(f"Could not generate tokens for tenant: {tenant_id}; exception: {e}")
-            self.service_tokens[tenant_id] = {'access_token': tokens.access_token,
+            self.service_tokens[tenant_id] = {'access_token': self.add_claims_to_token(tokens.access_token),
                                               'refresh_token': tokens.refresh_token}
 
     def validate_token(self, token):
@@ -567,36 +600,44 @@ class Tapis(object):
         except Exception as e:
             raise errors.TokenInvalidError("Invalid Tapis token.")
 
+
+    def add_claims_to_token(self, token):
+        """
+        Adds claims and a callable expires_in() function to a Tapis token object.
+        :param tokne: (TapisResult) A TapisResult object returned using the t.tokens.create_token() method.
+        """
+        def _expires_in():
+            return access_token.expires_at - datetime.datetime.now(datetime.timezone.utc)
+
+        access_token = token
+        access_token.claims = self.validate_token(access_token.access_token)
+        access_token.original_ttl = access_token.expires_in
+        access_token.expires_at = datetime.datetime.fromtimestamp(access_token.claims['exp'],
+                                                                  datetime.timezone.utc)
+        access_token.expires_in = _expires_in
+        return access_token
+
     def set_access_token(self, token):
         """
         Set the access token to be used in this session.
         :param token: (TapisResult) A TapisResult object returned using the t.tokens.create_token() method.
         :return:
         """
-
-        def _expires_in():
-            return self.access_token.expires_at - datetime.datetime.now(datetime.timezone.utc)
-
-        self.access_token = token
         try:
-            self.access_token.claims = self.validate_token(self.access_token.access_token)
-            self.access_token.original_ttl = self.access_token.expires_in
-            self.access_token.expires_in = _expires_in
-            self.access_token.expires_at = datetime.datetime.fromtimestamp(self.access_token.claims['exp'],
-                                                                           datetime.timezone.utc)
+            self.access_token = self.add_claims_to_token(token=token)
         except:
             pass
 
-    def refresh_tokens(self ,tenant_id=None):
+
+    def refresh_tokens(self, tenant_id=None):
         """
         Use the refresh token on this client to get a new access and refresh token pair.
         """
-        if not self.refresh_token:
-            raise errors.TapyClientConfigurationError(msg="No refresh token found.")
         if self.account_type == 'service':
             return self.refresh_service_tokens(tenant_id)
-        else:
-            return self.refresh_user_tokens()
+        if not self.refresh_token:
+            raise errors.TapyClientConfigurationError(msg="No refresh token found.")
+        return self.refresh_user_tokens()
 
     def refresh_user_tokens(self):
         """
@@ -619,7 +660,7 @@ class Tapis(object):
         """
         refresh_token = self.service_tokens[tenant_id]['refresh_token'].refresh_token
         tokens = self.tokens.refresh_token(refresh_token=refresh_token)
-        self.service_tokens[tenant_id]['access_token'] = tokens.refresh_token
+        self.service_tokens[tenant_id]['access_token'] = self.add_claims_to_token(tokens.access_token)
         self.service_tokens[tenant_id]['refresh_token'] = tokens.refresh_token
 
     def set_refresh_token(self, token):
@@ -948,6 +989,7 @@ class Operation(object):
                 pass
         # finally, look a username on the tapis client itself
         if not user:
+            print(f"no user object, returning username on the tapis_client...")
             user = self.tapis_client.username
         return user
 
@@ -973,7 +1015,7 @@ class Operation(object):
             # if we got a tenant_id, use it to look up the base_url:
             if tenant_id:
                 try:
-                    base_url = self.tapis_client.tenants.get_base_url_for_service_request(tenant_id, self.resource_name)
+                    base_url = self.tapis_client.tenant_cache.get_base_url_for_service_request(tenant_id, self.resource_name)
                 except:
                     pass
             # if all else fails, use the base_url for the client
@@ -1025,22 +1067,42 @@ class Operation(object):
 
         # construct the http headers -
         headers = {}
-        access_token = None
+
         # set the X-Tapis-Token header using the client
+        access_token = None
+        request_site_master_tenant_id = None    # used for looking up the service token
         if self.tapis_client.is_tapis_service and hasattr(self.tapis_client, 'service_tokens'):
+            # the tenant_id for the request could be a user tenant (e.g., "tacc" or "dev") but the
+            # service tokens are stored by master tenant, so we need to get the master tenant for the
+            # owning site of the tenant.
+            try:
+                request_tenant = self.tapis_client.tenant_cache.get_tenant_config(tenant_id=tenant_id)
+            except Exception as e:
+                raise errors.BaseTapyException(f"Could not lookup the request tenant for tenant {tenant_id}; e: {e}")
+            try:
+                request_site = request_tenant.site
+            except Exception as e:
+                raise errors.BaseTapyException(f"Could not lookup the request site for tenant {tenant_id}; e: {e}")
+            try:
+                request_site_master_tenant_id = request_site.site_master_tenant_id
+            except Exception as e:
+                raise errors.BaseTapyException(f"Could not lookup the request site_master tenant for "
+                                               f"tenant {tenant_id}; e: {e}")
+
             # service_tokens may be defined but still be empty dictionaries... this __call__ could be to get
             # the service's first set of tokens.
-            if tenant_id in self.tapis_client.service_tokens.keys() \
-                    and 'access_token' in self.tapis_client.service_tokens[tenant_id].keys():
+            if request_site_master_tenant_id in self.tapis_client.service_tokens.keys() \
+                    and 'access_token' in self.tapis_client.service_tokens[request_site_master_tenant_id].keys():
                 try:
-                    access_token = self.tapis_client.service_tokens[tenant_id]['access_token']
+                    access_token = self.tapis_client.service_tokens[request_site_master_tenant_id]['access_token']
                 except KeyError:
-                    raise errors.BaseTapyException(f"Did not find service tokens for tenant {tenant_id}; "
-                                                   f"service_tokens for tenant: {self.tapis_client.service_tokens[tenant_id]}")
+                    raise errors.BaseTapyException(f"Did not find service tokens for "
+                                                   f"tenant {request_site_master_tenant_id};")
             else:
                 pass
         elif self.tapis_client.get_access_jwt():
             access_token = self.tapis_client.get_access_jwt()
+        # if we got an access token, check if we need to refresh it
         if access_token:
             # check for a token about to expire in the next 5 seconds; assume by default we have a token with
             # plenty of time remaining.
@@ -1058,18 +1120,24 @@ class Operation(object):
                     pass
                 else:
                     try:
-                        self.tapis_client.refresh_tokens()
+                        self.tapis_client.refresh_tokens(tenant_id=request_site_master_tenant_id)
                     except:
                         # for now, if we get an error trying to refresh the tokens,s, we ignore it and try the
                         # request anyway.
                         pass
-            if self.tapis_client.is_tapis_service:
-                try:
-                    jwt = self.tapis_client.service_tokens[tenant_id]['access_token'].access_token
-                except KeyError:
-                    raise errors.BaseTapyException(f"No service tokens found for tenant {tenant_id} after refresh.")
-            else:
-                jwt = self.tapis_client.get_access_jwt()
+        # we may have refreshed the token, so we get it one more time --
+        if self.tapis_client.is_tapis_service \
+                and hasattr(self.tapis_client, 'service_tokens') \
+                and request_site_master_tenant_id \
+                and 'access_token' in self.tapis_client.service_tokens[request_site_master_tenant_id].keys():
+            try:
+                jwt = self.tapis_client.service_tokens[request_site_master_tenant_id]['access_token'].access_token
+            except Exception as e:
+                msg = f"Could not get JWT from access_token; e: {e}"
+                raise errors.BaseTapyException(msg)
+        else:
+            jwt = self.tapis_client.get_access_jwt()
+        if jwt:
             headers = {'X-Tapis-Token':  jwt}
 
         # the X-Tapis-Tenant and X-Tapis-Username headers must be set when the token represents a service account.
@@ -1078,11 +1146,10 @@ class Operation(object):
         if self.tapis_client.is_tapis_service:
             if tenant_id:
                 headers['X-Tapis-Tenant'] = tenant_id
-        # similarly for username, we first look in the request content:
-        if self.tapis_client.is_tapis_service:
-            x_isername = self.determine_user_for_service_request(**kwargs)
-        if self.tapis_client.x_username:
-            headers['X-Tapis-User'] = self.tapis_client.x_username
+            # similarly for username, we first look in the request content:
+            x_username = self.determine_user_for_service_request(**kwargs)
+            if x_username:
+                headers['X-Tapis-User'] = x_username
 
         # allow arbitrary headers to be passed in via the special "headers" kwarg -
         try:
@@ -1095,29 +1162,43 @@ class Operation(object):
         data = None
         # these are the list of allowable request body content types; ex., 'application/json'.
         if hasattr(self.op_desc.request_body, 'content') and hasattr(self.op_desc.request_body.content, 'keys'):
-            if 'application/json' in self.op_desc.request_body.content.keys() \
-                    or '*/*' in self.op_desc.request_body.content.keys():
-                headers['Content-Type'] = 'application/json'
-                required_fields = self.op_desc.request_body.content['application/json'].schema.required
-                data = {}
-                # if the request body has no defined properties, look for a single "request_body" parameter.
-                if self.op_desc.request_body.content['application/json'].schema.properties == {}:
-                    # choice of "request_body" is arbitrary, as the property name is not provided by the openapi spec in this case
+            # Before we go by spec headers, we first check if the user specifed any because the 
+            # 'else' would overwrite those if the spec specifies one of the following content-types.
+            # In the case of custom headers, we just pass whatever in (Christian's decision, probably 
+            # should improve).
+            if 'Content-Type' in headers:
+                data = kwargs['request_body']
+            # We go back to the old methods if there is not custom headers.
+            # Note: If an operation has many possible content-types, this just kind of breaks because
+            # all of the ifs return as true. e.g. sendMessage in Actors w/json, binary, and strings.
+            # In that case, these are more of order of which headers should be used. e.g. JSON headers
+            # would be used primarily unless it's not in the operation spec.
+            else:
+                if 'application/json' in self.op_desc.request_body.content.keys():
+                    headers['Content-Type'] = 'application/json'
+                    required_fields = self.op_desc.request_body.content['application/json'].schema.required
+                    data = {}
+                    # if the request body has no defined properties, look for a single "request_body" parameter.
+                    if self.op_desc.request_body.content['application/json'].schema.properties == {}:
+                        # choice of "request_body" is arbitrary, as the property name is not provided by the
+                        # openapi spec in this case
+                        data = kwargs['request_body']
+                    else:
+                        # otherwise, the request body has defined properties, so look for each one in the function kwargs
+                        for p_name, p_desc in self.op_desc.request_body.content['application/json'].schema.properties.items():
+                            if p_name in kwargs:
+                                data[p_name] = kwargs[p_name]
+                            elif p_name in required_fields:
+                                raise errors.InvalidInputError(msg=f'{p_name} is a required argument.')
+                        # serialize data before passing it to the request
+                    data = json.dumps(data)
+                elif 'application/octet-stream' in self.op_desc.request_body.content.keys():
+                    headers['Content-Type'] = 'application/octet-stream'
                     data = kwargs['request_body']
-                else:
-                    # otherwise, the request body has defined properties, so look for each one in the function kwargs
-                    for p_name, p_desc in \
-                            self.op_desc.request_body.content['application/json'].schema.properties.items():
-                        if p_name in kwargs:
-                            data[p_name] = kwargs[p_name]
-                        elif p_name in required_fields:
-                            raise errors.InvalidInputError(msg=f'{p_name} is a required argument.')
-                    # serialize data before passing it to the request
-                data = json.dumps(data)
-            if 'multipart/form-data' in self.op_desc.request_body.content.keys():
-                # todo - iterate over parts in self.op_desc.request_body.content['multipart/form-data'].schema.properties
-                raise NotImplementedError
-        # todo - handle other body content types..
+                elif 'multipart/form-data' in self.op_desc.request_body.content.keys():
+                    # todo - iterate over parts in self.op_desc.request_body.content['multipart/form-data'].schema.properties
+                    raise NotImplementedError
+                # todo - handle other body content types..
 
         # create a prepared request -
         # cf., https://requests.kennethreitz.org/en/master/user/advanced/#request-and-response-objects
@@ -1174,13 +1255,15 @@ class Operation(object):
         # resp.headers is a case-insensitive dict
         resp_content_type = resp.headers.get('content-type')
         if hasattr(resp_content_type, 'lower') and resp_content_type.lower() == 'application/json':
-            json_content = resp.json()
             # get the Tapis result object which could be a JSON object or list.
             try:
+                json_content = resp.json()
                 result = json_content.get('result')
             except Exception as e:
                 # some Tapis APIs, such as the Meta API, do not return "result" objects and/or do not return the
                 # standard Tapis stanzas but rather "raw" JSON documents
+                if debug:
+                    return resp.content, debug_data
                 return resp.content
             if result or result == [] or result == {}:
                 # if it is a list we should return a list of TapisResult objects:
